;; Teiresias tutorial part 2

;; This section of the tutorial introduces
;; some more complicated concepts

;; ==================================================
;; UNION TYPES

(if (read) 1 'Teiresias)

;; If an expression can produce values of more than
;; one type, Teiresias produces a "union" type,
;; of the form
;;     (+ type1 type2 ... typeN)

;; --------------------------------------------------
;; APPROXIMATION

(define identity-1 (lambda (x) x))

(identity-1 '1)

(identity-1 "Hallo")

;; Teiresias "merges" the arguments from different
;; calls to itentity-1

;; --------------------------------------------------
;; CHECKS

(car nil)

;; Teiresias finds primitives whose arguments have an
;; unexpected type, 
;; and highlights these primitives in red.

;; --------------------------------------------------
;; APPROXIMATION

(define identity-2 (lambda (x) x))

(identity-2 '1)

(car (identity-2 (cons 1 2)))

;; However, even if you know the argument is always ok,
;; Teiresias may not always be smart enough 
;; to figure it out.

;; --------------------------------------------------
;; IF-SPLITTING

(define to-string 
  (lambda (x)
    (if (number? x)
        (number->string x)
        (symbol->string x))))

(to-string 4)

(to-string 'Teiresias)

;; Teiresias is fairly smart about predicates 
;; such as number?, pair?, etc.
;; It knows that x must be a number in the then branch,
;; and cannot be a number in the else branch.

;; --------------------------------------------------
;; FLOW-SENSITIVE

(define copy-pair
  (lambda (p)
    (let* ([a (car p)]
           [d (cdr p)])
      (cons a d))))

(copy-pair (cons 1 2))
(copy-pair 'not-a-pair)

;; Teiresias knows that if the operation car succeeds, 
;; then p must be bound to a pair, 
;; and therefore no check is needed on the cdr.

;; --------------------------------------------------
;; RECURSIVE TYPES

(define make-tree
  (lambda (depth)
    (if (zero? depth)
        'leaf
        (cons (make-tree (sub1 depth))
              (make-tree (sub1 depth))))))

(define tree (make-tree 1024))

;; The type of tree is 
;;          (rec ([Y1 (+ sym (cons Y1 Y1))]) Y1)
;; This means that tree has type Y1,
;; where Y1 is either a symbol, 
;;             or is a pair containing two sub-trees, 
;;                each of type Y1.

;; --------------------------------------------------
;; LIST TYPE

(define make-list
  (lambda (n)
    (if (zero? n)
        '()
        (cons n (make-list (sub1 n))))))

(define my-list (make-list 1024))

;; The type of my-list is "really":
;;          (rec ([Y1 (+ nil (cons num Y1))]) Y1)
;; But this type is so common we abbreviate it as:
;;          (list num)

;; --------------------------------------------------
;; IF-SPLITTING (not if mutable)

(define to-string2
  (lambda (x)
    (if (number? x)
        (begin 
          (set! x 'food)
          (number->string x))
        (symbol->string x))))

(to-string2 4)
(to-string2 'Teiresias)

;; Teiresias doesn't do if-splitting on assigned variables

(define to-string3
  (lambda (x)
    (set! x x)
    (if (number? x)
        (number->string x)
        (symbol->string x))))

(to-string3 4)
(to-string3 'Teiresias)

;; --------------------------------------------------
;; FLOW-SENSITIVE (not if mutable)

(define copy-pair
  (lambda (p)
    (set! p p)
    (let* ([a (car p)]
           [d (cdr p)])
      (cons a d))))

(copy-pair (cons 1 2))
(copy-pair 'not-a-pair)

;; Because p is assigned, Teiresias can't track the
;; information that p must be a pair
;; if the operation car succeeds.

;; --------------------------------------------------
;; DEFINE-STRUCTURE

(define-structure (node left right))

(define n (make-node 1 2))
(node-left n)

(set-node-right! n 'symbol)
(node-right n)

;; if-splitting
(let ([x (if (read) n 4)])
  (if (node? x)
      (node-left n)
      'error))

;; flow-sensitive
(let ([x (if (read) n 4)])
  (+ (node-left x)
     (node-right x)))

;; ==================================================
              
