
;
; Some functions that are commonly provided in other Scheme implementations
;

(require-library "constant.ss")
(require-library "sfunctor.ss")
(require-library "function.ss")

(define-signature mzlib:compat^
  1+ 1- #%1+ #%1-
  >=? <=? >? <? =?
  real-time
  flush-output-port
  bound?
  sort
  gentemp
  cons? atom?
  putprop getprop
  new-cafe)

(define-sigfunctor (mzlib:compat@ mzlib:compat^)
  (import mzlib:function^)
  
  (define 1+ add1)
  (define 1- sub1)
  (define #%1+ #%add1)
  (define #%1- #%sub1)
  
  (define =? =)
  (define <? <)
  (define >? >)
  (define <=? <)
  (define >=? >)

  (define atom? (lambda (v) (not (pair? v))))

  (define cons? pair?)

  (define gentemp gensym)

  (define sort ; Chez argument order
    (lambda (less-than? l)
      (mzlib:function^:quicksort l less-than?)))

  (define bound? defined?)

  (define flush-output-port flush-output)

  (define real-time current-milliseconds)

  (define getprop (void))
  (define putprop (void))
  (let ([table (make-hash-table)])
    (letrec ([gp
	      (case-lambda
	       [(k prop) (gp k prop #f)]
	       [(k prop def)
		(let ([al (hash-table-get table k (lambda () #f))])
		  (if al
		      (let ([v (assq prop al)])
			(if v
			    (cdr v)
			    def))
		      def))])]
	     [pp
	      (lambda (k prop nv)
		(let ([al (hash-table-get table k (lambda () '()))])
		  (let ([v (assq prop al)])
		    (if v
			(set-cdr! v nv)
			(hash-table-put! table k (cons (cons prop nv) al))))))])
      (set! getprop gp)
      (set! putprop pp)))

  ; Chez's new-cafe
  (define new-cafe 
    (letrec ([nc
	      (case-lambda
	       [() (nc (current-eval))]
	       [(eval)
		(catch escape
		       (let ([orig-exit (exit-handler)]
			     [orig-eval (current-eval)])
			 (dynamic-wind
			  (lambda ()
			    (current-eval eval)
			    (exit-handler
			     (lambda (v) (escape v))))
			  read-eval-print-loop
			  (lambda ()
			    (current-eval orig-eval)
			    (exit-handler orig-exit)))))])])
      nc))
  )

(let ()
  (define-compound-sigfunctor (linked@ mzlib:compat^)
    (import)
    (with (mzlib:compat@ mzlib:function@)
	  (mzlib:function@))
    (export mzlib:compat@))
  
  (open-mzlib-functor linked@ (mzlib:compat@)))

(mzlib-constant 1+)
(mzlib-constant 1-)
(mzlib-constant #%1+)
(mzlib-constant #%1-)
(mzlib-constant >=?)
(mzlib-constant <=?)
(mzlib-constant >?)
(mzlib-constant <?)
(mzlib-constant =?)
(mzlib-constant real-time)
(mzlib-constant flush-output-port)
(mzlib-constant bound?)
(mzlib-constant sort)
(mzlib-constant gentemp)
(mzlib-constant cons?)
(mzlib-constant atom?)
(mzlib-constant putprop)
(mzlib-constant getprop)
(mzlib-constant new-cafe)

(define-macro defmacro
  (lambda (name args . body)
    `(define-macro ,name (lambda ,args ,@body))))
(constant defmacro)

(define-macro letmacro
  (lambda (name args mbody . body)
    `(let-macro ,name (lambda ,args ,mbody)
		,body)))
(constant letmacro)

(define-macro make-class* 
 (lambda (supers . rest)
  (let ([first (if (pair? rest) (car rest) #f)])
    (if (and (pair? first)
	     (pair? (car first)))
	; Old "new" syntax
	`(make-class* ,supers ,@(car rest) ,@(cdr rest))
	(let-values ([(ivars init)
		      (let loop ([l rest])
			(cond
			 [(null? l)
			  (values '()
				  (if (null? supers)
				      'void
				      `(lambda args
					 ,@(map (lambda (x)
						  (list
						   'apply
						   (string->symbol
						    (string-append (symbol->string (car x))
								   "-init"))
						   'args))
						supers))))]
			 [(not (pair? l))
			  (error 'make-class* "bad syntax (improper form)")]
			 [(and (pair? (car l))
			       (member (caar l) '(public
						  private
						  local
						  inherit
						  rename
						  share
						  inherit-from
						  rename-from
						  share-from)))
			  (let-values ([(ivars init) (loop (cdr l))])
				      (values (cons (car l) ivars) init))]
			 [(not (null? (cdr l)))
			  (error 'make-class* "bad syntax (multiple init expressions)")]
			 [else
			  (values '() (car l))]))])
		    (let ([gen (gensym)])
		      `(class* ,supers ,gen
			       ,@ivars
			       (private
				,@(map (lambda (s)
					 `(,(string->symbol
					     (string-append "apply-" 
							    (symbol->string (car s))
							    "-init"))
					   (lambda args
					     (#%apply ,(string->symbol
							(string-append (symbol->string 
									(car s))
								       "-init"))
						      (let loop ([l args])
							(if (#%null? (cdr l))
							    (#%car l)
							    (#%cons (#%car l) 
								    (loop (#%cdr l)))))))))
				       supers))
			       (sequence
				 (apply ,init
					,gen)))))))))
(constant make-class*)

(define-macro make-class 
  (lambda (super . rest)
    `(make-class* ((super ,super)) ,@rest)))
(constant make-class)
