;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Gann Bierner                                -=)
;; gbierner@owlnet.rice.edu                     (=-             \ O
;;                                                _O              \_-)---
;; File: share.ss                              (=-_/             /\
;;                                                 /\    
;;
;; This file contains code which formats an expression to show all sharing
;; within it. Call `print-convert' to convert a value.  
;; It takes 2 optional arguments.  The first is a boolean value, just-circular.
;; If true, sharing will only be shown for circularity.  The default is #f
;; where all sharing is shown.  The first argument to share:print-all is, of
;; course, the expression to convert.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require-library "constant.ss")
(require-library "sfunctor.ss")
(require-library "strings.ss")
(require-library "function.ss")

(define-signature mzlib:print-convert^
  print-convert
  print-convert-expr
  build-share
  get-shared)

(define-signature mzlib:print-convert-hooks^
  print-convert-hook
  build-share-hook
  build-share-name-hook)

(define-sigfunctor (mzlib:print-convert@ mzlib:print-convert^)
  (import mzlib:string^ mzlib:function^ mzlib:print-convert-hooks^)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; the value stored in the hash table.  Contains the name (a number
;; unless we are in donkey and it already has a name) and whether
;; or not it is shared in the expr.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-struct share-info (name shared?))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; share-hash is the hash-table containing info on what cons cells
;; of the expression are shared.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; sometimes you want to go ahead and start displaying a shared
;; expression rather than just showing its name.  For instance, in
;; the shared list, you want (shared ((<1> (list 1 2))... not
;; (shared ((<1> <1>) ...
;; expand-shared? controls this
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-struct convert-share-info (share-hash expand-shared?))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; builds the hash table
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define build-share
  (lambda (expr)
    (letrec*
	([share-cnt 0]
	 [share-hash (make-hash-table)]
	 [csi (make-convert-share-info share-hash #f)]
	 [hash
	  (lambda (obj)
	    (let ([name (mzlib:print-convert-hooks^:build-share-name-hook obj)])
	      (hash-table-put! share-hash obj
			       (make-share-info (if name (car name) share-cnt) #f)))
	    (set! share-cnt (add1 share-cnt)))]
	 [rehash
	  (lambda (val)
	    (if val (set-share-info-shared?! val #t))
	    val)]
	 [build-sub
	  (lambda (expr)
	    (rehash (hash-table-get share-hash expr (lambda () (hash expr) #f))))]
	 [build
	  (lambda (expr)
	    (cond
	      [(or (procedure? expr) (regexp? expr) (promise? expr) (string? expr)) 
	       (build-sub expr)]
	      [(box? expr) (unless (build-sub expr) (build (unbox expr)))]
	      [(pair? expr) (unless (build-sub expr) 
				    (build (car expr))
				    (build (cdr expr)))]
	      [(vector? expr) (unless (build-sub expr) (for-each build (vector->list expr)))]
	      [(struct? expr) (unless (build-sub expr)
				      (for-each build (vector->list (struct->vector expr))))]
	      [else (mzlib:print-convert-hooks^:build-share-hook expr build)]))])
	(build expr)
	csi)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; creates a distinctive symbol out of a name (usually just a number)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define map-share-name
  (lambda (name)
    (string->symbol
     (string-append "<" (mzlib:string^:expr->string name) ">"))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; prints an expression given that it has already been hashed.  This
;; does not include the list of shared items.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define print-convert-expr
  (lambda (csi expr)
    (letrec* ([share-hash (convert-share-info-share-hash csi)]
	      [shared?
	       (lambda (info)
		 (if info
		     (share-info-shared? info)
		     #f))]
	     [find-hash
	      (lambda (expr)
		(hash-table-get share-hash expr (lambda () #f)))]
	     [make-arg-list
	      (lambda (x)
		(let ([answer
		       (let loop ([x (if (struct? x) (arity-at-least-value x) x)] [a '()])
			 (if (zero? x)
			     a
			     (loop (sub1 x)
				   (cons (string->symbol
					  (string-append "a" (number->string x))) a))))])
		  (if (struct? x)
		      (if (null? x)
			  'args
			  (append answer 'args))
		      answer)))]
	     [print
	      (lambda (expr)
		(let ([es (convert-share-info-expand-shared? csi)]
		      [lookup (find-hash expr)])
		  (set-convert-share-info-expand-shared?! csi #f)
		  (if (and lookup (not es) (share-info-shared? lookup))
		      (map-share-name (share-info-name lookup))
		      (cond
			 [(null? expr) 'null]
			 [(weak-box? expr) `(make-weak-box ,(weak-box-value expr))]
			 [(box? expr) `(box ,(print (unbox expr)))]
			 [(vector? expr) (cons 'vector (map print (vector->list expr)))]
			 [(symbol? expr) `(quote ,expr)]
			 [(string? expr) expr]
			 [(primitive? expr) (string->symbol (primitive-name expr))]
			 [(procedure? expr) `(lambda ,(make-arg-list (arity expr)) ...)]
			 [(regexp? expr) '(regexp ...)]
			 [(void? expr) '(void)]
			 [(promise? expr) '(delay ...)]
			 [(pair? expr)
			  (if (and (shared? (find-hash (cdr expr))))
			      `(cons ,(print (car expr)) ,(print (cdr expr)))
			      (let ([rest (print (cdr expr))]
				    [first (print (car expr))])
				(if (or (eq? rest 'null)
					(and (pair? rest)
					     (eq? (car rest) 'list)))
				    (append `(list ,first)
					    (if (eq? 'null rest)
						'()
						(cdr rest)))
				    `(cons ,first ,rest))))]
			 [(struct? expr) (let ([name (symbol->string
						   (vector-ref (struct->vector expr) 0))])
					(cons (string->symbol
					       (string-append
						"make-" (substring name
								   (string-length "struct:")
								   (string-length name))))
					      (map print (cdr (vector->list
							       (struct->vector expr))))))]
			 [else (mzlib:print-convert-hooks^:print-convert-hook 
				expr print)]))))])
	      (print expr))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; these functions get the list of shared items.  If just-circular is
;; true, then it will modify the hash table so that the only shared
;; items are those that are circular.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define get-shared-helper
  (lambda (csi)
    (let ([shared '()]
	  [share-hash (convert-share-info-share-hash csi)])
      (hash-table-for-each share-hash
			   (lambda (key val)
			     (if (share-info-shared? val)
				 (set! shared (cons (list key val) shared)))))
      (map (lambda (s)
	     (set-convert-share-info-expand-shared?! csi #t)
	     (list (cadr s)
		   (map-share-name (share-info-name (cadr s)))
		   (print-convert-expr csi (car s))))
	   shared))))

(define get-shared
  (case-lambda 
   [(csi) (get-shared csi #f)]
   [(csi just-circular)
    (let ([shared (get-shared-helper csi)])
      (when just-circular
	    (for-each (lambda (x)
			(if (not (member* (cadr x) (caddr x)))
			    (set-share-info-shared?! (car x) #f)))			     
		      shared)
	    (set! shared (get-shared-helper csi)))
      (map cdr (mzlib:function^:quicksort 
		shared
		(lambda (x y) 
		  (string<? (mzlib:string^:expr->string (share-info-name (car x)))
			    (mzlib:string^:expr->string (share-info-name (car y))))))))]))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; helper function for determining if an item is circular.  In the
;; shared list: (shared ((<1> (list 1 2)) (<2> (list <2> 2 3)))), you
;; can tell by doing a member* of the first item on the second.  In this
;; case, the second item in the shared list is circular because <2> appears
;; in the value
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define member*
  (lambda (a l)
    (cond [(or (not (pair? l)) (null? l)) #f]
	  [(eq? a (car l)) #t]
	  [else (or (member* a (car l)) (member* a (cdr l)))])))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; takes an expression and completely converts it to show sharing
;; (or if just-circular, just circularity) and special forms.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define print-convert
  (case-lambda 
   [(expr) (print-convert expr #f)]
   [(expr just-circ)
    (let* ([csi (build-share expr)])
      (let ([shared (get-shared csi just-circ)])
	(if (null? shared)
	    (print-convert-expr csi expr)
	    (list 'shared shared (print-convert-expr csi expr)))))]))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; some tests to see if everything is going ok.  This isn't complete
;; but it's pretty good.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define test
  (lambda ()
    (letrec ([test
              (lambda (before right-answer)
                (let ([found (print-convert before)])
                  (unless (equal? right-answer found)
                    (printf "error: found: ~s expected ~s~n"
                            found right-answer))))]
	     [share-list
	      (let* ([a (list 1 2 3)]
		     [b (list 1 a (cdr a))])
		(set-car! b b)
		(append b (list (list 2 3))))]
	     [share-vector
	      (let ([v (vector 1 share-list (cdr share-list))])
		(vector-set! v 0 v)
		v)]
             [tests
              (list (cons "abc" "abc")
                    (cons 8 8)
                    (cons 'a '(quote a))
                    (cons '(1) '(list 1))
                    (cons #10(0) '(vector 0 0 0 0 0 0 0 0 0 0))
                    (cons (delay 1) '(delay ...))
                    (cons (let-struct a (a) (make-a 3)) '(make-a 3))
                    (cons (box 3) '(box 3))
                    (cons () 'null)
                    (cons add1 'add1)
                    (cons (void) '(void))
                    (cons (make-weak-box 12) '(make-weak-box 12))
                    (cons (regexp "1") '(regexp ...))
                    (cons (lambda (x) x) '(lambda (a1) ...))
                    (cons (lambda x x)  '(lambda args ...))
                    (cons (lambda (a b . c) a) '(lambda (a1 a2 . args) ...))
                    (cons (list 'a (box (list ())) (cons 1 ()))
                          '(list (quote a) (box (list null)) (list 1)))
		    (cons '#0=(#0#) '(shared ((<0> (list <0>))) <0>))
		    (cons '#2=(1 . #2#) '(shared ((<0> (cons 1 <0>))) <0>))
		    (cons share-list
			  '(shared
			    ((<1> (list <1> <3> <4>))
			     (<3> (cons 1 <4>))
			     (<4> (list 2 3)))
			    (list <1> <3> <4> (list 2 3))))
		    (cons share-vector
			  '(shared
			    ((<0> (vector <0> (cons <2> <8>) <8>))
			     (<2> (list <2> <4> <5>))
			     (<4> (cons 1 <5>))
			     (<5> (list 2 3))
			     (<8> (list <4> <5> (list 2 3))))
			    <0>))
                    )])
      (for-each (lambda (x)
                  (printf "testing: ~s~n" (car x))
                  (test (car x) (cdr x)))
                tests))))

)


(define-sigfunctor (mzlib:print-convert-hooks@ mzlib:print-convert-hooks^)
  (import)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; these are hooks that allow you to add more possibilities to the
;; types of expressions that are specially hashed or printed.
;; They both take an expression and a function to call if you want
;; to recur.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  (define build-share-hook (lambda (x b) 'no-share))
  (define build-share-name-hook  (lambda (x) #f))
  (define print-convert-hook (lambda (x p) x)))


#|

(let ()
  (define-compound-sigfunctor (linked@ mzlib:print-convert^)
    (import)
    (with (mzlib:print-convert@ mzlib:string@ mzlib:function@
				mzlib:print-convert-hooks@)
	  (mzlib:string@)
	  (mzlib:function@)
	  (mzlib:print-convert-hooks@))
    (export mzlib:print-convert@))

  (invoke-open-functor (sigfunctor->functor linked@) (mzlib:print-convert@)))

|#

(let ()
  (define-compound-sigfunctor (linked@ mzlib:print-convert^)
    (import)
    (with (mzlib:print-convert@ mzlib:string@ mzlib:function@
				mzlib:print-convert-hooks@)
	  (mzlib:string@)
	  (mzlib:function@)
	  (mzlib:print-convert-hooks@))
    (export mzlib:print-convert@ mzlib:print-convert-hooks@))

  (open-mzlib-functor linked@ #f))

(mzlib-constant print-convert)

