(define aries:error-report (box 'no-error-report))

(define-struct aries:unbound-error (id location))
(define-struct aries:app-error (location))

(let ((c-e-h (current-exception-handler)))
  (current-exception-handler
    (lambda (exn)
      (let ((report (unbox aries:error-report)))
	(cond
	  ((and (exn:variable? exn) (aries:unbound-error? report))
	    (zodiac:report-dynamic-error
	      (format "Unbound identifier: ~a"
		(aries:unbound-error-id report))
	      (aries:unbound-error-location report)
	      '()))
	  ((and (exn:application? exn) (aries:app-error? report))
	    (zodiac:report-dynamic-error
	      (exn-message exn)
	      (aries:app-error-location report)
	      '()))
	  (else
	    (c-e-h exn)))))))

; ----------------------------------------------------------------------

(define aries:primitive-list
  (list
   'list 'null?  'cons?  'atom?  'list?  'boolean?  'symbol?  'char? 'not
   'vector?  'structure?
   'pair?  'string?  'procedure?  'number?  'complex?  'real?
   'rational?  'integer?  'cons 'eq?  'eqv?  'equal?  'add1 'sub1 'zero?
   'exact?  'inexact?  '+ '- '* '/ 'odd?  'even?  'positive?
   'negative?  'symbol->string 'string->symbol 'abs 'length 'reverse
   'quotient 'remainder 'modulo 'gcd 'lcm 'numerator 'denominator 'floor
   'ceiling 'truncate 'round 'rationalize 'exp 'log 'sin 'cos 'tan 'asin
   'acos 'atan 'sqrt 'expt 'make-rectangular 'make-polar 'real-part
   'imag-part 'magnitude 'angle 'exact->inexact 'inexact->exact
   'number->string 'string->number 'memq 'memv 'member 'assq 'assv 'assoc
   'char=?  'char<?  'char<=?  'char>?  'char>=?  'char-ci=?  'char-ci<?
   'char-ci<=?  'char-ci>?  'char-ci>=?  'char-alphabetic?
   'char-numeric?  'char-whitespace?  'char-upper-case?
   'char-lower-case?  'char->integer 'integer->char 'char-upcase
   'char-downcase 'make-string 'string 'string-length 'string-ref
   'string=? 'string<? 'string>? 'string<=? 'string>=?
   'string-ci=? 'string-ci<? 'string-ci>? 'string-ci<=? 'string-ci>=?
   'list->string 'string->list 'string-append 'substring 'string-set!  
   'string-fill! 'string-copy 
   'vector 'make-vector 'vector-length 'vector-set! 'vector->list
   'list->vector 'vector-fill!
   'apply 'map 'for-each 'force
   'call-with-current-continuation 'call/cc
   'call-with-input-file  'call-with-output-file
   'with-input-from-file  'with-output-to-file
   'open-input-file  'open-output-file 'close-input-port  'close-output-port
   'current-input-port  'current-output-port 'input-port?  'output-port?
   'read 'read-char 'peek-char 'char-ready?   'eof-object?
   'write 'display 'write-char 'newline 'load 'transcript-on 'transcript-off
   '< '<= '= '> '>= 'max 'min 'length 'append 'reverse
   'list-ref 'list-tail 'vector-ref 'set-car!  'set-cdr!  'car 'cdr 'caar
   'cadr 'cdar 'cddr 'caaar 'caadr 'cadar 'caddr 'cdaar 'cdadr 'cddar
   'cdddr 'caaaar 'caaadr 'caadar 'caaddr 'cadaar 'cadadr 'caddar 'cadddr
   'cdaaar 'cdaadr 'cdadar 'cdaddr 'cddaar 'cddadr 'cdddar 'cddddr))

(define aries:defined-list
  (box '()))

(define aries:annotate
  (let ((memq memq) (string->symbol string->symbol))
    (lambda (expr)
      (let ((location (zodiac:determine-start-location expr)))
	(cond
	  ((zodiac:lexical-varref? expr)
	    (zodiac:syntax-symbol-name (zodiac:lexical-varref-var expr)))
	  ((zodiac:top-level-varref? expr)
	    (zodiac:syntax-symbol-name (zodiac:top-level-varref-var expr)))
	  ((zodiac:unbound-varref? expr)
	    (let ((name (zodiac:syntax-symbol-name
			  (zodiac:unbound-varref-var expr))))
	      (if (or (memq name aries:primitive-list)
		    (memq name (unbox aries:defined-list)))
		name
		`(begin
		   (set-box! aries:error-report
		     (make-aries:unbound-error ',name ,location))
		   ,name))))

	  ((zodiac:app? expr)
	    (let ((fun (aries:annotate (zodiac:app-fun expr)))
		   (args (map aries:annotate (zodiac:app-args expr))))
	      (if (null? args)
		`(begin
		   (let ((fun ,fun))
		     (set-box! aries:error-report
		       (make-aries:app-error ,location))
		     (fun)))
		`(,fun                ; assumes l->r order of evaluation!
		   ,@(let loop
		        ((args args))
		       (if (null? (cdr args))
			 `((let ((result ,(car args)))
			     (set-box! aries:error-report
			       (make-aries:app-error ,location))
			     result))
			 (cons (car args) (loop (cdr args)))))))))

	  ((zodiac:delay-form? expr)
	    `(delay ,(aries:annotate (zodiac:delay-form-expr expr))))
	  ((zodiac:if-form? expr)
	    `(if ,(aries:annotate (zodiac:if-form-test expr))
	       ,(aries:annotate (zodiac:if-form-then expr))
	       ,(aries:annotate (zodiac:if-form-else expr))))
	  ((zodiac:lambda-form? expr)
	    `(lambda ,(zodiac:improper-map (lambda (x)
					      (zodiac:syntax-symbol-name
						(zodiac:bound-var x)))
			 (zodiac:lambda-form-args expr))
	       ,(aries:annotate (zodiac:lambda-form-body expr))))

	  ((zodiac:set!-form? expr)
	    (let ((s (zodiac:syntax-symbol-name (zodiac:set!-form-var expr))))
	      (if (not (memq s (unbox aries:defined-list)))
		(set-box! aries:defined-list
		  (cons s (unbox aries:defined-list))))
	      `(set! ,s
		 ,(aries:annotate (zodiac:set!-form-val expr)))))
	  ((zodiac:time-form? expr)
	    `(time ,(aries:annotate (zodiac:time-form-expr expr))))
	  ((zodiac:quote-form? expr)
	    (let ((v (zodiac:strip* (zodiac:quote-form-expr expr))))
	      `(quote ,v)))
	  ((zodiac:begin-form? expr)
	    `(begin ,(aries:annotate (zodiac:begin-form-first expr))
	       ,(aries:annotate (zodiac:begin-form-rest expr))))
	  ((zodiac:let-form? expr)
	    (let ((bindings (zodiac:syntax-map (lambda (var val)
						 (list (zodiac:bound-var var)
						   (aries:annotate val)))
			      (zodiac:let-form-vars expr)
			      (zodiac:let-form-vals expr))))
	      `(let ,bindings
		 ,(aries:annotate (zodiac:let-form-body expr)))))
	  ((zodiac:letrec-form? expr)
	    (let ((bindings (zodiac:syntax-map (lambda (var val)
						 (list (zodiac:bound-var var)
						   (aries:annotate val)))
			      (zodiac:letrec-form-vars expr)
			      (zodiac:letrec-form-vals expr))))
	      `(letrec ,bindings
		 ,(aries:annotate (zodiac:letrec-form-body expr)))))
	  ((zodiac:define-form? expr)
	    (let ((s (zodiac:syntax-symbol-name
		       (zodiac:bound-var
			 (zodiac:define-form-name expr)))))
	      (if (not (memq s (unbox aries:defined-list)))
		(set-box! aries:defined-list
		  (cons s (unbox aries:defined-list))))
	      `(define ,s
		 ,(aries:annotate (zodiac:define-form-value expr)))))

	  (else (error 'aries:annotate "shouldn't get here ~s" expr)))))))

(define aries:transform
  (let ((input-port? input-port?) (open-input-file open-input-file)
	 (current-eval current-eval) (list list))
    (lambda (filename/port file-key interactive?)
      (let ([reader (zodiac:read
		      (if (input-port? filename/port)
			filename/port
			(open-input-file filename/port))
		      (make-zodiac:location
			(zodiac:location-line zodiac:default-scan-offset)
			(zodiac:location-column zodiac:default-scan-offset)
			(zodiac:location-offset zodiac:default-scan-offset)
			file-key))]
	     [eva-lu-ator (current-eval)])
	(let read-loop
	  ([eval-values (list (zodiac:void))])
	  (let ((expr (reader)))
	    (if (zodiac:eof? expr)
	      (if interactive?
		(zodiac:report-result file-key eval-values)
		(zodiac:report-loaded file-key))
	      (zodiac:call-with-values
		(lambda ()
		  (let* ((expanded (zodiac:expand expr))
			  (annotated (aries:annotate expanded))
			  (stripped (zodiac:strip* annotated)))
		    (eva-lu-ator stripped)))
		(lambda vals
		  (read-loop vals))))))))))

(define aries:
  (let ((current-input-port current-input-port))
    (lambda ()
      (aries:transform (current-input-port) 'console #t))))

(define aries:print-transform
  (let ((input-port? input-port?) (open-input-file open-input-file)
	 (current-eval current-eval) (list list))
    (lambda (filename/port file-key interactive?)
      (let ([reader (zodiac:read
		      (if (input-port? filename/port)
			filename/port
			(open-input-file filename/port))
		      (make-zodiac:location
			(zodiac:location-line zodiac:default-scan-offset)
			(zodiac:location-column zodiac:default-scan-offset)
			(zodiac:location-offset zodiac:default-scan-offset)
			file-key))]
	     [eva-lu-ator pretty-print])
	(let read-loop
	  ([eval-values (list (zodiac:void))])
	  (let ((expr (reader)))
	    (if (zodiac:eof? expr)
	      (if interactive?
		(zodiac:report-result file-key eval-values)
		(zodiac:report-loaded file-key))
	      (zodiac:call-with-values
		(lambda ()
		  (let* ((expanded (zodiac:expand expr))
			  (annotated (aries:annotate expanded))
			  (stripped (zodiac:strip* annotated)))
		    (eva-lu-ator stripped)))
		(lambda vals
		  (read-loop vals))))))))))

(define aries:print
  (let ((current-input-port current-input-port))
    (lambda ()
      (aries:print-transform (current-input-port) 'console #t))))

(if zodiac:print-on-startup
  (printf "Mars installed~n"))
