; Pretty-printer for MzScheme
;  Handles structures, cycles, and graphs
;
; Procedures:
;
;  (pretty-print v) - pretty-prints v
;  (pretty-print v port) - pretty-prints v to port
;  (pretty-print v port c) - pretty-prints v to port using c columns; if
;                            c is #f, (pretty-print-columns) is used
;  (pretty-print v port c graph?) - pretty-prints v to port using c columns
;                                   and use graph? instead of (print-graph)
;  (pretty-print v port c graph? struct?) - pretty-prints v to port using 
;                                   c columns, using graph? instead of 
;                                   (print-graph) and struct? instead of 
;                                   (print-struct)
;
;  (pretty-print-columns) - returns the default number of columns
;  (pretty-print-columns c) - sets the default number of columns
;    initial setting: 79
;
;  (pretty-print-handler v) - pretty-prints v if v is not #<void>
;
; TO INSTALL this pretty-printer into a MzScheme's read-eval-print loop,
;  load this file and evaluate:
;      (current-print pretty-print-handler)

; Originally:
 ;"genwrite.scm" generic write used by pp.scm
 ;;copyright (c) 1991, marc feeley

;; Matthew Flatt
;; Modified original for MrEd Spring/95
;; Added check for cyclic structures 11/9/95
;; Better (correct) graph printing, support boxes and structures 11/26/95

(define pretty-print #f)
(define pretty-print-columns #f)
(define pretty-print-handler #f)

(let () ; Begin package

(define print-def-cols 79)
(define -pretty-print-columns
  (case-lambda
   (() print-def-cols)
   ((v) (set! print-def-cols v))))

(define -pretty-print
  (case-lambda 
   ([obj port width print-graph? print-struct?]
    (generic-write obj #f 
		   (if width width (pretty-print-columns))
		   (lambda (s) (display s port) #t)
		   print-graph? print-struct?)
    (void))
   ([obj port width print-graph?] 
    (-pretty-print obj port width print-graph? (print-struct)))
   ([obj port width] 
    (-pretty-print obj port width (print-graph) (print-struct)))
   ([obj port] (-pretty-print obj port print-def-cols))
   ([obj] (-pretty-print obj (current-output-port) print-def-cols))))

(define (generic-write obj display? width output print-graph? print-struct?)

  (define table (make-hash-table)) ; Hash table for looking for loops

  (define-struct mark (str def))

  (define found-cycle
    (or print-graph?
	(let loop ([obj obj])
	  (and (or (vector? obj)
		   (pair? obj)
		   (box? obj)
		   (and (struct? obj) print-struct?))
	       (or (hash-table-get table obj (lambda () #f))
		   (begin
		     (hash-table-put! table obj #t)
		     (let ([cycle
			    (cond
			     [(vector? obj)
			      (ormap loop (vector->list obj))]
			     [(pair? obj)
			      (or (loop (car obj))
				  (loop (cdr obj)))]
			     [(box? obj) (loop (unbox obj))]
			     [(struct? obj)
			      (ormap loop 
				     (vector->list (struct->vector obj)))])])
		       (hash-table-remove! table obj)
		       cycle)))))))

  (if found-cycle
      (let loop ([obj obj])
	(if (or (vector? obj)
		(pair? obj)
		(box? obj)
		(and (struct? obj) print-struct?))
	    ; A little confusing: use #t for not-found
	    (let ([p (hash-table-get table obj (lambda () #t))])
	      (when (not (mark? p))
		    (if p
			(begin
			  (hash-table-put! table obj #f)
			  (cond
			   [(vector? obj)
			    (loop (vector->list obj))]
			   [(pair? obj)
			    (loop (car obj))
			    (loop (cdr obj))]
			   [(box? obj) (loop (unbox obj))]
			   [(struct? obj)
			    (for-each loop 
				      (vector->list (struct->vector obj)))]))
			(begin
			  (hash-table-put! table obj 
					   (make-mark #f (box #f))))))))))

  (define cycle-counter 0)

  (define found (if found-cycle
		    table 
		    #f))

 (let generic-write ([obj obj] [display? display?] 
		     [width width] [output output]
		     [def-box (box #t)])

  (define (read-macro? l)
    (define (length1? l) (and (pair? l) (null? (cdr l))))
    (let ((head (car l)) (tail (cdr l)))
      (case head
        ((quote quasiquote unquote unquote-splicing) (length1? tail))
        (else                                        #f))))

  (define (read-macro-body l)
    (cadr l))

  (define (read-macro-prefix l)
    (let ((head (car l)) (tail (cdr l)))
      (case head
        ((quote)            "'")
        ((quasiquote)       "`")
        ((unquote)          ",")
        ((unquote-splicing) ",@"))))

  (define (out str col)
    (and col (output str) (+ col (string-length str))))

  (define expr-found
    (lambda (ref col)
      (let ([n cycle-counter])
	(set! cycle-counter (add1 cycle-counter))
	(set-mark-str! ref 
		       (string-append "#"
				      (number->string n)
				      "#"))
	(set-mark-def! ref def-box)
	(out (string-append "#"
			    (number->string n)
			    "=")
	     col))))
  
    (define check-expr-found
      (lambda (obj check? col c-k d-k n-k)
	(let ([ref (and check? 
			found
			(hash-table-get found obj (lambda () #f)))])
	   (if (and ref (unbox (mark-def ref)))
	       (if c-k
		   (c-k (mark-str ref) col)
		   (out (mark-str ref) col))
	       (if (and ref d-k)
		   (d-k col)
		   (let ([col (if ref
				  (expr-found ref col)
				  col)])
		     (n-k col)))))))

  (define (wr obj col)

    (define (wr-expr expr col)
      (if (read-macro? expr)
        (wr (read-macro-body expr) (out (read-macro-prefix expr) col))
        (wr-lst expr col #t)))

    (define (wr-lst l col check?)
      (if (pair? l)
          (check-expr-found 
	   l check? col
	   #f #f
	   (lambda (col)
	     (let loop ((l (cdr l)) (col (wr (car l) (out "(" col))))
	       (check-expr-found
		l (and check? (pair? l)) col
		(lambda (s col) (out ")" (out s (out " . " col))))
		(lambda (col)
		  (wr-lst l (out " . " col) check?))
		(lambda (col)
		  (and col
		       (cond 
			((pair? l) 
			 (loop (cdr l) (wr (car l) (out " " col))))
			((null? l) (out ")" col))
			(else
			 (out ")" (wr l (out " . " col)))))))))))
	  (out "()" col)))

    (cond ((pair? obj)        (wr-expr obj col))
          ((null? obj)        (wr-lst obj col #f))
          ((vector? obj)      (check-expr-found
			       obj #t col
			       #f #f
			       (lambda (col)
				 (wr-lst (vector->list obj) 
					 (out "#" col) #f))))
	  ((box? obj)         (check-expr-found
			       obj #t col
			       #f #f
			       (lambda (col)
				 (wr (unbox obj) (out "#&" col)))))
	  ((struct? obj)      (if print-struct?
				  (check-expr-found
				   obj #t col
				   #f #f
				   (lambda (col)
				     (wr-lst (vector->list 
					      (struct->vector obj))
					     (out "#" col) #f)))
				  (out
				   (let ([p (open-output-string)]
					 [p-s (print-struct)])
				     (when p-s
				       (print-struct #f))
				     (display obj p)
				     (when p-s
					(print-struct p-s))
				     (get-output-string p))
				   col)))

          ((boolean? obj)     (out (if obj "#t" "#f") col))
          ((number? obj)      (out (number->string obj) col))
	  ; Let symbol get printed by default case to get proper quoting
          ; ((symbol? obj)      (out (symbol->string obj) col))
          ((string? obj)      (if display?
                                (out obj col)
                                (let loop ((i 0) (j 0) (col (out "\"" col)))
                                  (if (and col (< j (string-length obj)))
                                    (let ((c (string-ref obj j)))
                                      (if (or (char=? c #\\)
                                              (char=? c #\"))
                                        (loop j
                                              (+ j 1)
                                              (out "\\"
                                                   (out (substring obj i j)
                                                        col)))
                                        (loop i (+ j 1) col)))
                                    (out "\""
                                         (out (substring obj i j) col))))))
          ((char? obj)        (if display?
                                (out (make-string 1 obj) col)
                                (out (case obj
                                       ((#\space)   "space")
                                       ((#\newline) "newline")
                                       (else        (make-string 1 obj)))
                                     (out "#\\" col))))
	  (else (out (let ([p (open-output-string)])
		       (display obj p)
		       (get-output-string p))
		     col))))

  (define (pp obj col)

    (define (spaces n col)
      (if (> n 0)
        (if (> n 7)
          (spaces (- n 8) (out "        " col))
          (out (substring "        " 0 n) col))
        col))

    (define (indent to col)
      (and col
           (if (< to col)
             (and (out (make-string 1 #\newline) col) (spaces to 0))
             (spaces (- to col) col))))

    (define (pr obj col extra pp-pair)
      ; may have to split on multiple lines
      (let* ([can-multi (or (pair? obj) (vector? obj) 
			    (box? obj) (and (struct? obj) print-struct?))]
	     [ref (if can-multi
		      (and found (hash-table-get found obj (lambda () #f)))
		      #f)])
	(if (and can-multi
		 (or (not ref) (not (unbox (mark-def ref)))))
	    (let ((result '())
		  (new-def-box (box #t))
		  (left (min (+ (- (- width col) extra) 1) max-expr-width)))
	      (generic-write obj display? #f
	       (lambda (str)
		 (set! result (cons str result))
		 (set! left (- left (string-length str)))
		 (> left 0))
	       new-def-box)
	      (if (> left 0) ; all can be printed on one line
		  (out (reverse-string-append result) col)
		  (begin
		    (set-box! new-def-box #f)
		    (let ([col
			   (if ref
			       (expr-found ref col)
			       col)])
		      (cond
		       [(pair? obj) (pp-pair obj col extra)]
		       [(vector? obj)
			(pp-list (vector->list obj) 
				 (out "#" col) extra pp-expr #f)]
		       [(struct? obj)
			(pp-list (vector->list (struct->vector obj))
				 (out "#" col) extra pp-expr #f)]
		       [(box? obj)
			(pr (unbox obj) (out "#&" col) extra pp-pair)])))))
	    (wr obj col))))

    (define (pp-expr expr col extra)
      (if (read-macro? expr)
        (pr (read-macro-body expr)
            (out (read-macro-prefix expr) col)
            extra
            pp-expr)
        (let ((head (car expr)))
          (if (symbol? head)
            (let ((proc (style head)))
              (if proc
                (proc expr col extra)
                (if (> (string-length (symbol->string head))
                       max-call-head-width)
                  (pp-general expr col extra #f #f #f pp-expr)
                  (pp-call expr col extra pp-expr))))
            (pp-list expr col extra pp-expr #t)))))

    ; (head item1
    ;       item2
    ;       item3)
    (define (pp-call expr col extra pp-item)
      (let ((col* (wr (car expr) (out "(" col))))
        (and col
             (pp-down (cdr expr) col* (+ col* 1) extra pp-item #t #t))))

    ; (item1
    ;  item2
    ;  item3)
    (define (pp-list l col extra pp-item check?)
      (let ((col (out "(" col)))
        (pp-down l col col extra pp-item #f check?)))

    (define (pp-down l col1 col2 extra pp-item check-first? check-rest?)
      (let loop ((l l) (col col1) (check? check-first?))
        (and col
	     (check-expr-found
	      l (and check? (pair? l)) col
	      (lambda (s col) 
		(out ")" (out s (indent col2 (out "." (indent col2 col))))))
	      (lambda (col)
		(out ")" (pr l (indent col2 (out "." (indent col2 col)))
			     extra pp-item)))
	      (lambda (col)
		(cond ((pair? l)
		       (let ((rest (cdr l)))
			 (let ((extra (if (null? rest) (+ extra 1) 0)))
			   (loop rest
				 (pr (car l) (indent col2 col)
				     extra pp-item)
				 check-rest?))))
		      ((null? l)
		       (out ")" col))
		      (else
		       (out ")"
			    (pr l
				(indent col2 (out "." (indent col2 col)))
				(+ extra 1)
				pp-item)))))))))

    (define (pp-general expr col extra named? pp-1 pp-2 pp-3)

      (define (tail1 rest col1 col2 col3)
        (if (and pp-1 (pair? rest))
          (let* ((val1 (car rest))
                 (rest (cdr rest))
                 (extra (if (null? rest) (+ extra 1) 0)))
            (tail2 rest col1 (pr val1 (indent col3 col2) extra pp-1) col3))
          (tail2 rest col1 col2 col3)))

      (define (tail2 rest col1 col2 col3)
        (if (and pp-2 (pair? rest))
          (let* ((val1 (car rest))
                 (rest (cdr rest))
                 (extra (if (null? rest) (+ extra 1) 0)))
            (tail3 rest col1 (pr val1 (indent col3 col2) extra pp-2)))
          (tail3 rest col1 col2)))

      (define (tail3 rest col1 col2)
        (pp-down rest col2 col1 extra pp-3 #f #t))

      (let* ((head (car expr))
             (rest (cdr expr))
             (col* (wr head (out "(" col))))
        (if (and named? (pair? rest))
          (let* ((name (car rest))
                 (rest (cdr rest))
                 (col** (wr name (out " " col*))))
            (tail1 rest (+ col indent-general) col** (+ col** 1)))
          (tail1 rest (+ col indent-general) col* (+ col* 1)))))

    (define (pp-expr-list l col extra)
      (pp-list l col extra pp-expr #t))

    (define (pp-lambda expr col extra)
      (pp-general expr col extra #f pp-expr-list #f pp-expr))

    (define (pp-if expr col extra)
      (pp-general expr col extra #f pp-expr #f pp-expr))

    (define (pp-cond expr col extra)
      (pp-call expr col extra pp-expr-list))

    (define (pp-case expr col extra)
      (pp-general expr col extra #f pp-expr #f pp-expr-list))

    (define (pp-and expr col extra)
      (pp-call expr col extra pp-expr))

    (define (pp-let expr col extra)
      (let* ((rest (cdr expr))
             (named? (and (pair? rest) (symbol? (car rest)))))
        (pp-general expr col extra named? pp-expr-list #f pp-expr)))

    (define (pp-begin expr col extra)
      (pp-general expr col extra #f #f #f pp-expr))

    (define (pp-do expr col extra)
      (pp-general expr col extra #f pp-expr-list pp-expr-list pp-expr))

    ; define formatting style (change these to suit your style)

    (define indent-general 2)

    (define max-call-head-width 5)

    (define max-expr-width 50)

    (define (style head)
      (case head
        ((lambda let* letrec define) pp-lambda)
        ((if set!)                   pp-if)
        ((cond)                      pp-cond)
        ((case)                      pp-case)
        ((and or)                    pp-and)
        ((let)                       pp-let)
        ((begin)                     pp-begin)
        ((do)                        pp-do)
        (else                        #f)))

    (pr obj col 0 pp-expr))

  (if width
    (out (make-string 1 #\newline) (pp obj 0))
    (wr obj 0))))

; (reverse-string-append l) = (apply string-append (reverse l))

(define (reverse-string-append l)

  (define (rev-string-append l i)
    (if (pair? l)
      (let* ((str (car l))
             (len (string-length str))
             (result (rev-string-append (cdr l) (+ i len))))
        (let loop ((j 0) (k (- (- (string-length result) i) len)))
          (if (< j len)
            (begin
              (string-set! result k (string-ref str j))
              (loop (+ j 1) (+ k 1)))
            result)))
      (make-string i)))

  (rev-string-append l 0))

; Export these procedures
(set! pretty-print -pretty-print)
(set! pretty-print-columns -pretty-print-columns)

(set! pretty-print-handler
  (lambda (v)
    (unless (void? v)
       (-pretty-print v))))

) ; end pretty-print package
